require "box2d"

Box2d = Core.class(Sprite)

function Box2d:init(debug)

	self.world					= b2.World.new(0, 20, true)
	self.debug					= debug or false
	self.framemod				= 1
	self.framerate				= 60
	self.frametime				= 1 / self.framerate
	self.frameaccum				= 0
	self.paused					= false
	self.levelObjectsLive		= {}
	self.levelObjectsStatic		= {}
	self.levelObjectsDead		= {}
	self.removingObjects		= {}
	self.objectGroups			= {}
	
	self.events					= {
		beforeStep	= {},
		afterStep	= {},
		afterInit	= {}
	}
	
	self.world:addEventListener(Event.PRE_SOLVE, self.eventHandler, self)
	self.world:addEventListener(Event.BEGIN_CONTACT, self.eventHandler, self)
	self.world:addEventListener(Event.END_CONTACT, self.eventHandler, self)
	self.world:addEventListener(Event.POST_SOLVE, self.eventHandler, self)
	
end

function Box2d:initWorld(levelSize, objects, classes, states, extend, interact)
	
	self.levelObjectsDef		= objects
	self.levelObjectsCls		= classes
	self.size					= levelSize
	self.states					= states
	self.extend					= extend
	self.interact				= interact
	
	--if not self.levelObjectsDef then return end
	--[[
	for k, object in pairs(self.levelObjectsDef) do
		
		local objId		= tostring(object.id)
		local deleted	= self.states.dead[objId]
		
		if deleted then
			
			table.insert(self.levelObjectsDead, deleted)
			
		else
			
			local model = self.levelObjectsCls[object.type]
			
			if model then
				
				local createdObject = model(object, self, self.interact)
				
				local savedState	= self.states.live[objId]
				local extendState	= self.extend[objId]
				
				if savedState then
					createdObject:setState(savedState)
				end
				
				if extendState then
					createdObject:setState(extendState)
				end
				
				self:addLiveObject(createdObject)
				
			else
				
				self:createStaticObject(object)
				
			end
			
		end
		
	end
	]]
	--if self.debug then
		--set up debug drawing
		local debugDraw = b2.DebugDraw.new()
		self.world:setDebugDraw(debugDraw)
		self:addChild(debugDraw)
		
	--end
	
	--create bounding walls outside the scene
	self:createStaticObject({
		id			= 0,
		x			= 0,
		y			= 0,
		width		= 0,
		height		= 0,
		type		= "levelBorder",
		shape		= "polyline",
		polyline	= {
			{x = 0, y = 0},
			{x = self.size.w, y = 0},
			{x = self.size.w, y = self.size.h},
			{x = 0, y = self.size.h},
			{x = 0, y = 0}
		},
		properties = {
			friction = 0
		}
	})
	
	--self:addLighting(interact.camera)
	--self:arrangeLiveObjects()
	
end

function Box2d:addLighting(camera)

	local objects = {}
	
	for k, object in pairs(self.levelObjectsStatic) do
		if object.ext and object.ext.shade and tonumber(object.ext.shade) == 1 then
			table.insert(objects, object)
		end
	end
	
	for k, object in pairs(self.levelObjectsLive) do
		if object.ext and object.ext.shade and tonumber(object.ext.shade) == 1 then
			table.insert(objects, object)
		end
	end
	
	self.lightning = DynamicLight.new(self, objects, camera, 1)
	
	for id, object in pairs(self.levelObjectsLive) do
		
		if object.type == "light" then
			self.lightning:addSource(object)
		end
		
	end
	
	self.lightning:prepareRender()
	self:addChild(self.lightning)

end

function Box2d:eventHandler(event)
	
	local eventType	= event:getType()
	local fixtureA	= event.fixtureA
	local fixtureB	= event.fixtureB
	local bodyA		= fixtureA:getBody()
	local bodyB		= fixtureB:getBody()
	
	if bodyB.object[eventType] then
		bodyB.object[eventType](bodyB.object, fixtureB, fixtureA, bodyA, event)
	end
	
	if bodyA.object[eventType] then
		bodyA.object[eventType](bodyA.object, fixtureA, fixtureB, bodyB, event)
	end
	
end

-- creating objects using shapes
function Box2d:createStaticObject(obj)
	
	local x			= obj.x + obj.width / 2
	local y			= obj.y + obj.height / 2
	local props		= obj.properties or {}
	local friction	= tonumber(props.friction) or 0.8

	local object = {}
	
	--create box2d physical object
	local body = self.world:createBody{type = b2.STATIC_BODY}
	
	body:setPosition(x,y)
	
	if(obj.rotation) then
		
		body:setAngle(obj.rotation * math.pi / 180)
		
	end
	
	--define type of body
	local bodyTypeHandler = {
		["polygon"]		= B2Shape.polygon,
		["polyline"]	= B2Shape.polyline,
		["ellipse"]		= B2Shape.ellipse,
		["rectangle"]	= B2Shape.rectangle,
	}
	
	local bShape	= bodyTypeHandler[obj.shape](B2Shape, obj)
	local fixture	= body:createFixture({shape = bShape, density = 0, friction = friction, restitution = 0})
	
	object.id		= obj.id
	object.type		= obj.type
	object.width	= obj.width
	object.height	= obj.height
	object.shape	= obj.shape
	object.vertices	= obj.polygon or obj.polyline
	object.ext		= obj.properties or {}
	
	object.body 	= body
	body.object		= object
	
	table.insert(self.levelObjectsStatic, object)
	
end

function Box2d:arrangeLiveObjects()
	
	table.sort(self.levelObjectsLive, function(a, b)
		
		local a = a.zIndex or 0
		local b = b.zIndex or 0
		
		return tonumber(a) < tonumber(b)
		
	end)
	
	for k, o in pairs(self.levelObjectsLive) do
		self:swapChildrenAt(self:getChildIndex(o), k)
	end
	
end

function Box2d:addLiveObject(object)
	
	table.insert(self.levelObjectsLive, object)
	
	if self.lightning and object.ext and object.ext.shade then
		self.lightning:addObject(object)
	end
	
	local zIndex = object.zIndex
	
	if not object.zIndex or object.zIndex > self:getNumChildren() then
		zIndex = self:getNumChildren() + 1
	end
	
	self:addChildAt(object, zIndex)
	
end

function Box2d:removeLiveObject(object)
	
	self.removingObjects[object] = true
	
end

function Box2d:getObjectGroup(name)
	
	if not self.objectGroups[name] then
		self.objectGroups[name] = {}
	else
		tableReset(self.objectGroups[name])
	end
	
	for i = 1, #self.levelObjectsLive do
		
		local object =  self.levelObjectsLive[i]
		
		if object.type == name then
			
			self.objectGroups[name][#self.objectGroups[name]+1] = object
			
		end
		
	end
	
	return self.objectGroups[name]
	
end

function Box2d:getState()
	
	local worldState = {live = {}, dead = {}}
	
	for i = 1, #self.levelObjectsLive do
		
		local object = self.levelObjectsLive[i]
		
		if object.getState then
			
			worldState.live[tostring(object.id)] = object:getState()
			
		end
		
	end
	
	for i = 1, #self.levelObjectsDead do
		
		local object = self.levelObjectsDead[i]
		
		worldState.dead[tostring(object.id)] = object
		
	end
	
	return worldState
	
end

function Box2d:registerEvent(type, func)
	
	table.insert(self.events[type], func)
	
end

function Box2d:invokeEvents(type)
	
	for i = 1, #self.events[type] do
		
		local event = self.events[type][i]
		
		event()
		
	end
	
	--tableReset(self.events[type])
	
end
--[[
function Box2d:onEnterFrame(e)
	
	if self.paused then return end
	
	self:invokeEvents("beforeStep")
	
	self.world:step(self.frametime, 8, 3)
	
	self:invokeEvents("afterStep")
	
	for i = #self.levelObjectsLive, 1, -1 do
		
		local object	= self.levelObjectsLive[i]
		local remove	= self.removingObjects[object]
		
		if object.body then
			
			if remove then
				
				if self.lightning and object.ext and object.ext.shade then
					self.lightning:delObject(object)
				end
				
				object:removeFromParent()
				self.world:destroyBody(object.body)
				
				table.remove(self.levelObjectsLive, i)
				table.removekey(self.removingObjects, object)
				
				if object.getState then
					table.insert(self.levelObjectsDead, {id = object.id})
				end
				
				object.body = nil
				
			elseif object.body:getType() > 0 and object.body:isAwake() then
				
				object:setPosition(object.body:getPosition())
				object:setRotation(^>object.body:getAngle())
				
			end
			
		end
		
	end
	
	self.lightning:renderShadows()
	
end
--]]
--[
function Box2d:onEnterFrame(e) --print("Box2d:onEnterFrame"..math.random())
	
	if self.paused then return end
	
	self:invokeEvents("beforeStep")
	
	--Frame skipping
	self.frameaccum = self.frameaccum + e.deltaTime / self.framemod
	
	while self.frameaccum >= self.frametime do
		
		self.frameaccum = self.frameaccum - self.frametime
		
		if self.frameaccum < self.frametime then
			
			for i = 1, #self.levelObjectsLive do
				
				local object	= self.levelObjectsLive[i]
				
				if object.body and object.body:getType() > 0 then
					
					local x, y	= object.body:getPosition()
					local a		= object.body:getAngle()
					
					if object.state then
						object.state.x = x
						object.state.y = y
						object.state.a = a
					end
					
				end
				
			end
			
		end
		
		self.world:step(self.frametime, 8, 3)
		
	end
	
	local d = self.frameaccum / self.frametime
	
	self:invokeEvents("afterStep")
	
	for i = #self.levelObjectsLive, 1, -1 do
		
		local object	= self.levelObjectsLive[i]
		local remove	= self.removingObjects[object]
		
		if object.body then
			
			if remove then
				
				if self.lightning and object.ext and object.ext.shade then
					self.lightning:delObject(object)
				end
				
				object:removeFromParent()
				self.world:destroyBody(object.body)
				
				table.remove(self.levelObjectsLive, i)
				table.removekey(self.removingObjects, object)
				
				if object.getState then
					table.insert(self.levelObjectsDead, {id = object.id})
				end
				
				object.body = nil
				
			elseif object.body:getType() > 0 and object.body:isAwake() then
				
				local x, y	= object.body:getPosition()
				local a		= object.body:getAngle()
				
				local md	= 1 - d
				
				if not object.state then
					object.state = {
						x = x,
						y = y,
						a = a,
					}
				end
				
				object:setPosition(
					x * d + object.state.x * md,
					y * d + object.state.y * md
				)
				
				object:setRotation(
					^>(a * d + object.state.a * md)
				)
				
				--object:setPosition(x, y)
				--object:setRotation(^>a)
				
			end
			
		end
		
	end
	
--	self.lightning:renderShadows()
	
end
--]]
function Box2d:setFrameMod(framemod)
	
	self.framemod = framemod
	
end

function Box2d:getFrameMod()
	
	return self.framemod
	
end

function Box2d:runWorld()
	
	self:invokeEvents("afterInit")
	
	self:addEventListener(Event.ENTER_FRAME, self.onEnterFrame, self)
	
end

function Box2d:pauseWorld()
	
	self.paused = not self.paused
	
end

function Box2d:stopWorld()
	
	self:removeEventListener(Event.ENTER_FRAME, self.onEnterFrame, self)
	
end

function Box2d:delayedCall(time, handler)
	
	local timer		= Timer.new(1)
	local leftTime	= time
	local deltaTime	= 0
	
	timer:addEventListener(Event.TIMER, function()
		
		deltaTime	= 1 / self.framemod
		leftTime	= leftTime - deltaTime
		
		if leftTime <= 0 then
			timer:stop()
			handler()
		end
		
	end)
	
	timer:start()	
	
end